#!/usr/bin/env python3

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Setup script for installing torchgeo."""

import setuptools

if __name__ == "__main__":
    setuptools.setup()
